# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
This cooker creates tarballs containing a bare .git directory,
that can be unpacked and cloned like any git repository.

It works in three steps:

1. Write objects one by one in :file:`.git/objects/`
2. Calls ``git repack`` to pack all these objects into git packfiles.
3. Creates a tarball of the resulting repository

It keeps a set of all written (or about-to-be-written) object hashes in memory
to avoid downloading and writing the same objects twice.

The first step is the most complex. When swh-graph is available, this roughly does
the following:

1. Find all the revisions and releases in the induced subgraph, adds them to
   todo-lists
2. Grab a batch from (release/revision/directory/content) todo-lists, and load them.
   Add directory and content objects they reference to the todo-list
3. If any todo-list is not empty, goto 1

When swh-graph is not available, steps 1 and 2 are merged, because revisions need
to be loaded in order to compute the subgraph.
"""

import concurrent
import datetime
import enum
import glob
import logging
import os.path
import re
import subprocess
import tarfile
import tempfile
from typing import Any, Dict, Iterable, List, NoReturn, Optional, Set
import zlib

import sentry_sdk

from swh.core.api.classes import stream_results_optional
from swh.model import git_objects
from swh.model.hashutil import hash_to_bytehex, hash_to_hex
from swh.model.model import (
    Directory,
    DirectoryEntry,
    Person,
    Release,
    ReleaseTargetType,
    Revision,
    RevisionType,
    Sha1Git,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
    TimestampWithTimezone,
)
from swh.model.swhids import CoreSWHID, ObjectType
from swh.objstorage.interface import objid_from_dict
from swh.storage.algos.revisions_walker import DFSRevisionsWalker
from swh.storage.algos.snapshot import snapshot_get_all_branches
from swh.vault.cookers.base import BaseVaultCooker
from swh.vault.to_disk import HIDDEN_MESSAGE, SKIPPED_MESSAGE

RELEASE_BATCH_SIZE = 10000
REVISION_BATCH_SIZE = 10000
DIRECTORY_BATCH_SIZE = 10000
CONTENT_BATCH_SIZE = 100


logger = logging.getLogger(__name__)


class RootObjectType(enum.Enum):
    DIRECTORY = "directory"
    REVISION = "revision"
    RELEASE = "release"
    SNAPSHOT = "snapshot"


def assert_never(value: NoReturn, msg) -> NoReturn:
    """mypy makes sure this function is never called, through exhaustive checking
    of ``value`` in the parent function.

    See https://mypy.readthedocs.io/en/latest/literal_types.html#exhaustive-checks
    for details.
    """
    assert False, msg


class GitBareCooker(BaseVaultCooker):
    BUNDLE_TYPE = "git_bare"
    SUPPORTED_OBJECT_TYPES = {ObjectType[obj_type.name] for obj_type in RootObjectType}

    use_fsck = True

    obj_type: RootObjectType

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.obj_type = RootObjectType[self.swhid.object_type.name]
        self.executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=self.thread_pool_size
        )

    def check_exists(self) -> bool:
        """Returns whether the root object is present in the archive."""
        if self.obj_type is RootObjectType.REVISION:
            return not list(self.storage.revision_missing([self.obj_id]))
        elif self.obj_type is RootObjectType.RELEASE:
            return not list(self.storage.release_missing([self.obj_id]))
        elif self.obj_type is RootObjectType.DIRECTORY:
            return not list(self.storage.directory_missing([self.obj_id]))
        elif self.obj_type is RootObjectType.SNAPSHOT:
            return not list(self.storage.snapshot_missing([self.obj_id]))
        else:
            assert_never(self.obj_type, f"Unexpected root object type: {self.obj_type}")

    def _push(self, stack: List[Sha1Git], obj_ids: Iterable[Sha1Git]) -> None:
        """Adds all the given ``obj_ids`` to the given ``stack``, unless they are
        already in ``self._seen``, and adds them to ``self._seen``."""
        assert not isinstance(obj_ids, bytes)
        revision_ids = [id_ for id_ in obj_ids if id_ not in self._seen]
        self._seen.update(revision_ids)
        stack.extend(revision_ids)

    def _pop(self, stack: List[Sha1Git], n: int) -> List[Sha1Git]:
        """Removes ``n`` object from the ``stack`` and returns them."""
        obj_ids = stack[-n:]
        stack[-n:] = []
        return obj_ids

    def prepare_bundle(self) -> None:
        """Main entry point. Initializes the state, creates the bundle, and
        sends it to the backend."""
        # Objects we will visit soon (aka. "todo-lists"):
        self._rel_stack: List[Sha1Git] = []
        self._rev_stack: List[Sha1Git] = []
        self._dir_stack: List[Sha1Git] = []
        self._cnt_stack: List[Sha1Git] = []

        # Set of objects already in any of the stacks:
        self._seen: Set[Sha1Git] = set()
        self._walker_state: Optional[Any] = None

        # Set of errors we expect git-fsck to raise at the end:
        self._expected_fsck_errors: Set[str] = set()

        with tempfile.TemporaryDirectory(prefix="swh-vault-gitbare-") as workdir:
            # Initialize a Git directory
            self.workdir = workdir
            self.gitdir = os.path.join(workdir, "clone.git")
            os.mkdir(self.gitdir)
            self.init_git()

            self.nb_loaded = 0

            # Add the root object to the stack of objects to visit
            self.push_subgraph(self.obj_type, self.obj_id)

            # Load and write all the objects to disk
            self.load_objects()

            self.backend.set_progress(
                self.BUNDLE_TYPE, self.swhid, "Writing references..."
            )

            # Write the root object as a ref (this step is skipped if it's a snapshot)
            # This must be done before repacking; git-repack ignores orphan objects.
            self.write_refs()

            self.backend.set_progress(
                self.BUNDLE_TYPE, self.swhid, "Checking content integrity"
            )

            if self.use_fsck:
                self.git_fsck()

            self.backend.set_progress(
                self.BUNDLE_TYPE, self.swhid, "Creating final bundle"
            )

            self.repack()

            self.write_archive()

            self.backend.set_progress(self.BUNDLE_TYPE, self.swhid, "Uploading bundle")

    def init_git(self) -> None:
        """Creates an empty :file:`.git` directory."""
        subprocess.run(["git", "-C", self.gitdir, "init", "--bare"], check=True)
        self.create_object_dirs()

        # Remove example hooks; they take ~40KB and we don't use them
        for filename in glob.glob(os.path.join(self.gitdir, "hooks", "*.sample")):
            os.unlink(filename)

    def create_object_dirs(self) -> None:
        """Creates all possible subdirectories of :file:`.git/objects/`"""
        # Create all possible dirs ahead of time, so we don't have to check for
        # existence every time.
        for byte in range(256):
            try:
                os.mkdir(os.path.join(self.gitdir, "objects", f"{byte:02x}"))
            except FileExistsError:
                pass

    def repack(self) -> None:
        """Moves all objects from :file:`.git/objects/` to a packfile."""
        try:
            subprocess.run(["git", "-C", self.gitdir, "repack", "-d"], check=True)
        except subprocess.CalledProcessError:
            logging.exception("git-repack failed with:")
            sentry_sdk.capture_exception()

        # Remove their non-packed originals
        subprocess.run(["git", "-C", self.gitdir, "prune-packed"], check=True)

    def git_fsck(self) -> None:
        """Runs git-fsck and ignores expected errors (eg. because of missing
        objects)."""
        proc = subprocess.run(
            ["git", "-C", self.gitdir, "fsck"],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env={"LANG": "C.utf8"},
        )

        # Split on newlines not followed by a space
        errors = re.split("\n(?! )", proc.stdout.decode())

        errors = [
            error for error in errors if error and not error.startswith("warning ")
        ]

        unexpected_errors = set(errors) - self._expected_fsck_errors
        if unexpected_errors:
            logging.error(
                "Unexpected errors from git-fsck after cooking %s: %s",
                self.swhid,
                "\n".join(sorted(unexpected_errors)),
            )

    def _make_stub_directory_revision(self, dir_id: Sha1Git) -> Sha1Git:
        author = Person.from_fullname(
            b"swh-vault, git-bare cooker <robot@softwareheritage.org>"
        )
        dt = datetime.datetime.now(tz=datetime.timezone.utc)
        dt = dt.replace(microsecond=0)  # not supported by git
        date = TimestampWithTimezone.from_datetime(dt)

        revision = Revision(
            author=author,
            committer=author,
            date=date,
            committer_date=date,
            message=b"Initial commit",
            type=RevisionType.GIT,
            directory=self.obj_id,
            synthetic=True,
        )
        self.write_revision_node(revision)

        return revision.id

    def write_refs(self, snapshot=None) -> None:
        """Writes all files in :file:`.git/refs/`.

        For non-snapshot objects, this is only ``master``."""
        refs: Dict[bytes, bytes]  # ref name -> target
        if self.obj_type is RootObjectType.DIRECTORY:
            # We need a synthetic revision pointing to the directory
            rev_id = self._make_stub_directory_revision(self.obj_id)

            refs = {b"refs/heads/master": hash_to_bytehex(rev_id)}
        elif self.obj_type is RootObjectType.REVISION:
            refs = {b"refs/heads/master": hash_to_bytehex(self.obj_id)}
        elif self.obj_type is RootObjectType.RELEASE:
            (release,) = self.storage.release_get([self.obj_id])
            assert release, self.obj_id

            if release.name and re.match(rb"^[a-zA-Z0-9_.-]+$", release.name):
                release_name = release.name
            else:
                release_name = b"release"

            refs = {
                b"refs/tags/" + release_name: hash_to_bytehex(self.obj_id),
            }

            if release.target_type.value == ReleaseTargetType.REVISION.value:
                # Not necessary, but makes it easier to browse
                refs[b"refs/heads/master"] = hash_to_bytehex(release.target)
            # TODO: synthesize a master branch for other target types

        elif self.obj_type is RootObjectType.SNAPSHOT:
            if snapshot is None:
                # refs were already written in a previous step
                return
            branches = []
            for branch_name, branch in snapshot.branches.items():
                if branch is None:
                    logging.error(
                        "%s has dangling branch: %r", snapshot.swhid(), branch_name
                    )
                else:
                    branches.append((branch_name, branch))
            refs = {
                branch_name: (
                    b"ref: " + branch.target
                    if branch.target_type == SnapshotTargetType.ALIAS
                    else hash_to_bytehex(branch.target)
                )
                for (branch_name, branch) in branches
            }
        else:
            assert_never(self.obj_type, f"Unexpected root object type: {self.obj_type}")

        for ref_name, ref_target in refs.items():
            path = os.path.join(self.gitdir.encode(), ref_name)
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "wb") as fd:
                fd.write(ref_target)

    def write_archive(self):
        """Creates the final .tar file."""
        with tarfile.TarFile(mode="w", fileobj=self.fileobj) as tf:
            tf.add(self.gitdir, arcname=f"{self.swhid}.git", recursive=True)

    def _obj_path(self, obj_id: Sha1Git):
        """Returns the absolute path of file (in :file:`.git/objects/`) that will
        contain the git object identified by the ``obj_id``."""
        return os.path.join(self.gitdir, self._obj_relative_path(obj_id))

    def _obj_relative_path(self, obj_id: Sha1Git):
        """Same as :meth:`_obj_path`, but relative."""
        obj_id_hex = hash_to_hex(obj_id)
        directory = obj_id_hex[0:2]
        filename = obj_id_hex[2:]
        return os.path.join("objects", directory, filename)

    def object_exists(self, obj_id: Sha1Git) -> bool:
        """Returns whether the object identified by the given ``obj_id`` was already
        written to a file in :file:`.git/object/`.

        This function ignores objects contained in a git pack."""
        return os.path.exists(self._obj_path(obj_id))

    def write_object(self, obj_id: Sha1Git, obj: bytes) -> bool:
        """Writes a git object on disk.

        Returns whether it was already written."""
        # Git requires objects to be zlib-compressed; but repacking decompresses and
        # removes them, so we don't need to compress them too much.
        data = zlib.compress(obj, level=1)

        with open(self._obj_path(obj_id), "wb") as fd:
            fd.write(data)
        return True

    def push_subgraph(self, obj_type: RootObjectType, obj_id) -> None:
        """Adds graph induced by the given ``obj_id`` without recursing through
        directories, to the todo-lists.

        If swh-graph is not available, this immediately loads revisions, as they
        need to be fetched in order to compute the subgraph, and fetching them
        immediately avoids duplicate fetches."""
        if self.obj_type is RootObjectType.REVISION:
            self.push_revision_subgraph(obj_id)
        elif self.obj_type is RootObjectType.DIRECTORY:
            self._push(self._dir_stack, [obj_id])
        elif self.obj_type is RootObjectType.SNAPSHOT:
            self.push_snapshot_subgraph(obj_id)
        elif self.obj_type is RootObjectType.RELEASE:
            self.push_releases_subgraphs([obj_id])
        else:
            assert_never(self.obj_type, f"Unexpected root object type: {self.obj_type}")

    def load_objects(self) -> None:
        """Repeatedly loads objects in the todo-lists, until all lists are empty."""

        futures = []
        while self._rel_stack or self._rev_stack or self._dir_stack or self._cnt_stack:
            nb_remaining = (
                len(self._rel_stack)
                + len(self._rev_stack)
                + len(self._dir_stack)
                + len(self._cnt_stack)
            )
            # We assume assume nb_remaining is a lower bound.
            # When the snapshot was loaded with swh-graph, this should be the exact
            # value, though.
            self.backend.set_progress(
                self.BUNDLE_TYPE,
                self.swhid,
                f"Processing... {self.nb_loaded} objects processed\n"
                f"Over {nb_remaining} remaining",
            )

            release_ids = self._pop(self._rel_stack, RELEASE_BATCH_SIZE)
            if release_ids:
                self.load_releases(release_ids)
                self.nb_loaded += len(release_ids)

            revision_ids = self._pop(self._rev_stack, REVISION_BATCH_SIZE)
            if revision_ids:
                self.load_revisions(revision_ids)
                self.nb_loaded += len(revision_ids)

            directory_ids = self._pop(self._dir_stack, DIRECTORY_BATCH_SIZE)
            if directory_ids:
                self.load_directories(directory_ids)
                self.nb_loaded += len(directory_ids)

            content_ids = self._pop(self._cnt_stack, CONTENT_BATCH_SIZE)
            if content_ids:
                futures += [
                    self.executor.submit(self.load_content, content_id)
                    for content_id in content_ids
                ]
                self.nb_loaded += len(content_ids)

        self.backend.set_progress(
            self.BUNDLE_TYPE,
            self.swhid,
            "Fetching contents bytes ...",
        )
        concurrent.futures.wait(futures)

    def push_revision_subgraph(self, obj_id: Sha1Git) -> None:
        """Fetches the graph of revisions induced by the given ``obj_id`` and adds
        them to ``self._rev_stack``.

        If swh-graph is not available, this requires fetching the revisions themselves,
        so they are directly loaded instead."""
        loaded_from_graph = False

        if self.graph:
            from swh.graph.http_client import GraphArgumentException

            # First, try to cook using swh-graph, as it is more efficient than
            # swh-storage for querying the history
            obj_swhid = CoreSWHID(
                object_type=ObjectType.REVISION,
                object_id=obj_id,
            )
            try:
                revision_ids = (
                    swhid.object_id
                    for swhid in map(
                        CoreSWHID.from_string,
                        self.graph.visit_nodes(str(obj_swhid), edges="rev:rev"),
                    )
                )
                self._push(self._rev_stack, revision_ids)
            except GraphArgumentException as e:
                logger.info(
                    "Revision %s not found in swh-graph, falling back to fetching "
                    "history using swh-storage. %s",
                    hash_to_hex(obj_id),
                    e.args[0],
                )
            else:
                loaded_from_graph = True

        if not loaded_from_graph:
            # If swh-graph is not available, or the revision is not yet in
            # swh-graph, fall back to self.storage.revision_log.
            # self.storage.revision_log also gives us the full revisions,
            # so we load them right now instead of just pushing them on the stack.
            walker = DFSRevisionsWalker(self.storage, obj_id, state=self._walker_state)
            for rev_d in walker:
                if isinstance(rev_d, Revision):
                    # TODO: Remove this conditional after swh-storage v3.0.0 is released
                    revision = rev_d
                else:
                    revision = Revision.from_dict(rev_d)
                self.write_revision_node(revision)
                self.nb_loaded += 1
                self._push(self._dir_stack, [revision.directory])
            # Save the state, so the next call to the walker won't return the same
            # revisions
            self._walker_state = walker.export_state()

    def push_snapshot_subgraph(self, obj_id: Sha1Git) -> None:
        """Fetches a snapshot and all its children, excluding directories and contents,
        and pushes them to the todo-lists.

        Also loads revisions if swh-graph is not available, see
        :meth:`push_revision_subgraph`."""
        loaded_from_graph = False

        if self.graph:
            revision_ids = []
            release_ids = []
            directory_ids = []
            content_ids = []

            from swh.graph.http_client import GraphArgumentException

            # First, try to cook using swh-graph, as it is more efficient than
            # swh-storage for querying the history
            obj_swhid = CoreSWHID(
                object_type=ObjectType.SNAPSHOT,
                object_id=obj_id,
            )
            try:
                swhids: Iterable[CoreSWHID] = map(
                    CoreSWHID.from_string,
                    self.graph.visit_nodes(str(obj_swhid), edges="snp:*,rel:*,rev:rev"),
                )
                for swhid in swhids:
                    if swhid.object_type is ObjectType.REVISION:
                        revision_ids.append(swhid.object_id)
                    elif swhid.object_type is ObjectType.RELEASE:
                        release_ids.append(swhid.object_id)
                    elif swhid.object_type is ObjectType.DIRECTORY:
                        directory_ids.append(swhid.object_id)
                    elif swhid.object_type is ObjectType.CONTENT:
                        content_ids.append(swhid.object_id)
                    elif swhid.object_type is ObjectType.SNAPSHOT:
                        assert (
                            swhid.object_id == obj_id
                        ), f"Snapshot {obj_id.hex()} references a different snapshot"
                    else:
                        assert_never(
                            swhid.object_type, f"Unexpected SWHID object type: {swhid}"
                        )
            except GraphArgumentException as e:
                logger.info(
                    "Snapshot %s not found in swh-graph, falling back to fetching "
                    "history for each branch. %s",
                    hash_to_hex(obj_id),
                    e.args[0],
                )
            else:
                self._push(self._rev_stack, revision_ids)
                self._push(self._rel_stack, release_ids)
                self._push(self._dir_stack, directory_ids)
                self._push(self._cnt_stack, content_ids)
                loaded_from_graph = True

        # TODO: when self.graph is available and supports edge labels, use it
        # directly to get branch names.
        snapshot: Optional[Snapshot] = snapshot_get_all_branches(self.storage, obj_id)
        assert snapshot, "Unknown snapshot"  # should have been caught by check_exists()
        for branch in snapshot.branches.values():
            if not loaded_from_graph:
                if branch is None:
                    logging.warning("Dangling branch: %r", branch)
                    continue
                assert isinstance(branch, SnapshotBranch)  # for mypy
                if branch.target_type is SnapshotTargetType.REVISION:
                    self.push_revision_subgraph(branch.target)
                elif branch.target_type is SnapshotTargetType.RELEASE:
                    self.push_releases_subgraphs([branch.target])
                elif branch.target_type is SnapshotTargetType.ALIAS:
                    # Nothing to do, this for loop also iterates on the target branch
                    # (if it exists)
                    pass
                elif branch.target_type is SnapshotTargetType.DIRECTORY:
                    self._push(self._dir_stack, [branch.target])
                elif branch.target_type is SnapshotTargetType.CONTENT:
                    self._push(self._cnt_stack, [branch.target])
                elif branch.target_type is SnapshotTargetType.SNAPSHOT:
                    if swhid.object_id != obj_id:
                        raise NotImplementedError(
                            f"{swhid} has a snapshot as a branch."
                        )
                else:
                    assert_never(
                        branch.target_type, f"Unexpected target type: {self.obj_type}"
                    )

        self.write_refs(snapshot=snapshot)

    def load_revisions(self, obj_ids: List[Sha1Git]) -> None:
        """Given a list of revision ids, loads these revisions and their directories;
        but not their parent revisions (ie. this is not recursive)."""
        ret: List[Optional[Revision]] = self.storage.revision_get(obj_ids)

        revisions: List[Revision] = list(filter(None, ret))
        if len(ret) != len(revisions):
            logger.error("Missing revision(s), ignoring them.")

        for revision in revisions:
            self.write_revision_node(revision)
        self._push(self._dir_stack, (rev.directory for rev in revisions))

    def write_revision_node(self, revision: Revision) -> bool:
        """Writes a revision object to disk"""
        git_object = revision.raw_manifest or git_objects.revision_git_object(revision)
        return self.write_object(revision.id, git_object)

    def load_releases(self, obj_ids: List[Sha1Git]) -> List[Release]:
        """Loads release objects, and returns them."""
        ret = self.storage.release_get(obj_ids)

        releases = list(filter(None, ret))
        if len(ret) != len(releases):
            logger.error("Missing release(s), ignoring them.")

        for release in releases:
            self.write_release_node(release)

        return releases

    def push_releases_subgraphs(self, obj_ids: List[Sha1Git]) -> None:
        """Given a list of release ids, loads these releases and adds their
        target to the list of objects to visit"""
        for release in self.load_releases(obj_ids):
            self.nb_loaded += 1
            assert release.target, "{release.swhid(}) has no target"
            if release.target_type is ReleaseTargetType.REVISION:
                self.push_revision_subgraph(release.target)
            elif release.target_type is ReleaseTargetType.DIRECTORY:
                self._push(self._dir_stack, [release.target])
            elif release.target_type is ReleaseTargetType.CONTENT:
                self._push(self._cnt_stack, [release.target])
            elif release.target_type is ReleaseTargetType.RELEASE:
                self.push_releases_subgraphs([release.target])
            elif release.target_type is ReleaseTargetType.SNAPSHOT:
                raise NotImplementedError(
                    f"{release.swhid()} targets a snapshot: {release.target!r}"
                )
            else:
                assert_never(
                    release.target_type,
                    f"Unexpected release target type: {release.target_type}",
                )

    def write_release_node(self, release: Release) -> bool:
        """Writes a release object to disk"""
        git_object = release.raw_manifest or git_objects.release_git_object(release)
        return self.write_object(release.id, git_object)

    def load_directories(self, obj_ids: List[Sha1Git]) -> None:
        if not obj_ids:
            return

        raw_manifests = self.storage.directory_get_raw_manifest(obj_ids)

        concurrent.futures.wait(
            self.executor.submit(self.load_directory, obj_id, raw_manifests.get(obj_id))
            for obj_id in obj_ids
        )

    def load_directory(self, obj_id: Sha1Git, raw_manifest: Optional[bytes]) -> None:
        # Load the directory
        entries_it: Optional[Iterable[DirectoryEntry]] = stream_results_optional(
            self.storage.directory_get_entries, obj_id
        )

        if entries_it is None:
            logger.error("Missing swh:1:dir:%s, ignoring.", hash_to_hex(obj_id))
            return

        directory = Directory(
            id=obj_id, entries=tuple(entries_it), raw_manifest=raw_manifest
        )
        git_object = raw_manifest or git_objects.directory_git_object(directory)
        self.write_object(obj_id, git_object)

        # Add children to the stack
        entry_loaders: Dict[str, Optional[List[Sha1Git]]] = {
            "file": self._cnt_stack,
            "dir": self._dir_stack,
            "rev": None,  # Do not include submodule targets (rejected by git-fsck)
        }
        for entry in directory.entries:
            stack = entry_loaders[entry.type]
            if stack is not None:
                self._push(stack, [entry.target])

    def load_content(self, obj_id: Sha1Git) -> None:
        # TODO: add support of filtered objects, somehow?
        # It's tricky, because, by definition, we can't write a git object with
        # the expected hash, so git-fsck *will* choke on it.
        content = self.storage.content_get([obj_id], "sha1_git")[0]

        if content is None:
            # FIXME: this may also happen for missing content
            self.write_content(obj_id, SKIPPED_MESSAGE)
            self._expect_mismatched_object_error(obj_id)
        elif content.status == "visible":
            hashes = objid_from_dict(content.hashes())
            if self.objstorage is None:
                datum = self.storage.content_get_data(hashes)
            else:
                datum = self.objstorage.get(hashes)

            if datum is None:
                logger.error(
                    "%s is visible, but is missing data. Skipping.", content.swhid()
                )
            else:
                self.write_content(content.sha1_git, datum)
        elif content.status == "hidden":
            self.write_content(obj_id, HIDDEN_MESSAGE)
            self._expect_mismatched_object_error(obj_id)
        elif content.status == "absent":
            assert False, f"content_get returned absent content {content.swhid()}"
        else:
            # TODO: When content.status will have type Literal, replace this with
            # assert_never
            assert False, f"{content.swhid} has status: {content.status!r}"

    def write_content(self, obj_id: Sha1Git, content: bytes) -> None:
        header = git_objects.git_object_header("blob", len(content))
        self.write_object(obj_id, header + content)

    def _expect_mismatched_object_error(self, obj_id):
        obj_id_hex = hash_to_hex(obj_id)
        obj_path = self._obj_relative_path(obj_id)

        # For Git < 2.21:
        self._expected_fsck_errors.add(
            f"error: sha1 mismatch for ./{obj_path} (expected {obj_id_hex})"
        )
        # For Git >= 2.21:
        self._expected_fsck_errors.add(
            f"error: hash mismatch for ./{obj_path} (expected {obj_id_hex})"
        )

        self._expected_fsck_errors.add(
            f"error: {obj_id_hex}: object corrupt or missing: ./{obj_path}"
        )
        self._expected_fsck_errors.add(f"missing blob {obj_id_hex}")



# Copyright (C) 2017-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import contextlib
import datetime
import re
import smtplib
from unittest.mock import MagicMock, patch

import attr
from backports.entry_points_selectable import entry_points as get_entry_points
import psycopg2
import pytest
import requests

from swh.core.api import APIError
from swh.core.sentry import init_sentry
from swh.model.model import Content
from swh.model.swhids import CoreSWHID
from swh.vault.exc import NotFoundExc
from swh.vault.tests.vault_testing import hash_content

SENTRY_DSN = "https://user@example.org/1234"


@contextlib.contextmanager
def mock_cooking(vault_backend):
    with patch.object(vault_backend, "_send_task") as mt:
        mt.return_value = 42
        with patch("swh.vault.backend.get_cooker_cls") as mg:
            mcc = MagicMock()
            mc = MagicMock()
            mg.return_value = mcc
            mcc.return_value = mc
            mc.check_exists.return_value = True

            yield {
                "_send_task": mt,
                "get_cooker_cls": mg,
                "cooker_cls": mcc,
                "cooker": mc,
            }


def assertTimestampAlmostNow(ts, tolerance_secs=1.0):  # noqa
    now = datetime.datetime.now(datetime.timezone.utc)
    creation_delta_secs = (ts - now).total_seconds()
    assert creation_delta_secs < tolerance_secs


def fake_cook(backend, bundle_type, result_content, sticky=False):
    swhid = Content.from_data(result_content).swhid()
    content, obj_id = hash_content(result_content)
    with mock_cooking(backend):
        backend.create_task(bundle_type, swhid, sticky)
    backend.cache.add(bundle_type, swhid, b"content")
    backend.set_status(bundle_type, swhid, "done")
    return swhid, content


def fail_cook(backend, bundle_type, swhid, failure_reason):
    with mock_cooking(backend):
        backend.create_task(bundle_type, swhid)
    backend.set_status(bundle_type, swhid, "failed")
    backend.set_progress(bundle_type, swhid, failure_reason)


TEST_TYPE = "git_bare"
TEST_SWHID = CoreSWHID.from_string("swh:1:rev:4a4b9771542143cf070386f86b4b92d42966bdbc")
TEST_PROGRESS = (
    "Mr. White, You're telling me you're cooking again? \N{ASTONISHED FACE} "
)
TEST_EMAIL = "ouiche@lorraine.fr"


@pytest.fixture
def swh_vault(swh_vault, sample_data):
    # make the vault's storage consistent with test data
    revision = attr.evolve(sample_data.revision, id=TEST_SWHID.object_id)
    swh_vault.storage.revision_add([revision])
    return swh_vault


def test_create_task_simple(swh_vault):
    with mock_cooking(swh_vault) as m:
        swh_vault.create_task(TEST_TYPE, TEST_SWHID)

    m["get_cooker_cls"].assert_called_once_with(TEST_TYPE, TEST_SWHID.object_type)

    args = m["cooker_cls"].call_args[0]
    assert args[0] == TEST_SWHID

    assert m["cooker"].check_exists.call_count == 1
    assert m["_send_task"].call_count == 1

    args = m["_send_task"].call_args[0]
    assert args[0] == TEST_TYPE
    assert args[1] == TEST_SWHID

    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    assert info["swhid"] == TEST_SWHID
    assert info["type"] == TEST_TYPE
    assert info["task_status"] == "new"
    assert info["task_id"] == 42

    assertTimestampAlmostNow(info["ts_created"])

    assert info["ts_done"] is None
    assert info["progress_msg"] is None


def test_create_fail_duplicate_task(swh_vault):
    with mock_cooking(swh_vault):
        swh_vault.create_task(TEST_TYPE, TEST_SWHID)
        with pytest.raises(psycopg2.IntegrityError):
            swh_vault.create_task(TEST_TYPE, TEST_SWHID)


def test_create_fail_nonexisting_object(swh_vault):
    with mock_cooking(swh_vault) as m:
        m["cooker"].check_exists.side_effect = ValueError("Nothing here.")
        with pytest.raises(ValueError):
            swh_vault.create_task(TEST_TYPE, TEST_SWHID)


def test_create_set_progress(swh_vault):
    with mock_cooking(swh_vault):
        swh_vault.create_task(TEST_TYPE, TEST_SWHID)

    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    assert info["progress_msg"] is None
    swh_vault.set_progress(TEST_TYPE, TEST_SWHID, TEST_PROGRESS)
    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    assert info["progress_msg"] == TEST_PROGRESS


def test_create_set_status(swh_vault):
    with mock_cooking(swh_vault):
        swh_vault.create_task(TEST_TYPE, TEST_SWHID)

    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    assert info["task_status"] == "new"
    assert info["ts_done"] is None

    swh_vault.set_status(TEST_TYPE, TEST_SWHID, "pending")
    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    assert info["task_status"] == "pending"
    assert info["ts_done"] is None

    swh_vault.set_status(TEST_TYPE, TEST_SWHID, "done")
    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    assert info["task_status"] == "done"
    assertTimestampAlmostNow(info["ts_done"])


def test_create_update_access_ts(swh_vault):
    with mock_cooking(swh_vault):
        swh_vault.create_task(TEST_TYPE, TEST_SWHID)

    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    access_ts_1 = info["ts_last_access"]
    assertTimestampAlmostNow(access_ts_1)

    swh_vault.update_access_ts(TEST_TYPE, TEST_SWHID)
    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    access_ts_2 = info["ts_last_access"]
    assertTimestampAlmostNow(access_ts_2)

    swh_vault.update_access_ts(TEST_TYPE, TEST_SWHID)
    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)

    access_ts_3 = info["ts_last_access"]
    assertTimestampAlmostNow(access_ts_3)

    assert access_ts_1 < access_ts_2
    assert access_ts_2 < access_ts_3


def test_create_scheduler_rpc_failure(swh_vault, mocker):
    mocker.patch.object(swh_vault, "_send_task").side_effect = APIError(
        "Could not connect to scheduler RPC service"
    )
    # exception should bubble up when requesting to cook a bundle
    with pytest.raises(APIError):
        cooking_info = swh_vault.cook(TEST_TYPE, TEST_SWHID)

    # once scheduler rpc issue fixed, task should be created
    with mock_cooking(swh_vault):
        cooking_info = swh_vault.cook(TEST_TYPE, TEST_SWHID)
        assert cooking_info["task_status"] == "new"


def test_cook_idempotent(swh_vault, sample_data):
    with mock_cooking(swh_vault):
        info1 = swh_vault.cook(TEST_TYPE, TEST_SWHID)
        info2 = swh_vault.cook(TEST_TYPE, TEST_SWHID)
        info3 = swh_vault.cook(TEST_TYPE, TEST_SWHID)
        assert info1 == info2
        assert info1 == info3


def test_cook_email_pending_done(swh_vault):
    with (
        mock_cooking(swh_vault),
        patch.object(swh_vault, "add_notif_email") as madd,
        patch.object(swh_vault, "send_notification") as msend,
    ):
        swh_vault.cook(TEST_TYPE, TEST_SWHID)
        madd.assert_not_called()
        msend.assert_not_called()

        madd.reset_mock()
        msend.reset_mock()

        swh_vault.cook(TEST_TYPE, TEST_SWHID, email=TEST_EMAIL)
        madd.assert_called_once_with(TEST_TYPE, TEST_SWHID, TEST_EMAIL)
        msend.assert_not_called()

        madd.reset_mock()
        msend.reset_mock()

        swh_vault.set_status(TEST_TYPE, TEST_SWHID, "done")
        swh_vault.cache.add(TEST_TYPE, TEST_SWHID, b"content")
        swh_vault.cook(TEST_TYPE, TEST_SWHID, email=TEST_EMAIL)
        msend.assert_called_once_with(None, TEST_EMAIL, TEST_TYPE, TEST_SWHID, "done")
        madd.assert_not_called()


def test_send_all_emails(swh_vault):
    with mock_cooking(swh_vault):
        emails = ("a@example.com", "billg@example.com", "test+42@example.org")
        for email in emails:
            swh_vault.cook(TEST_TYPE, TEST_SWHID, email=email)

    swh_vault.set_status(TEST_TYPE, TEST_SWHID, "done")

    with patch.object(swh_vault, "_smtp_send") as m:
        swh_vault.send_notif(TEST_TYPE, TEST_SWHID)

        sent_emails = {k[0][0] for k in m.call_args_list}
        assert {k["To"] for k in sent_emails} == set(emails)

        download_url = (
            "https://archive.softwareheritage.org/api/1/vault/"
            f"{TEST_TYPE.replace('_', '-')}/{str(TEST_SWHID)}/raw"
        )

        for e in sent_emails:
            assert "bot@softwareheritage.org" in e["From"]
            assert TEST_TYPE in e["Subject"]
            assert TEST_SWHID.object_id.hex()[:5] in e["Subject"]
            assert TEST_TYPE in str(e)
            assert download_url in str(e)
            assert TEST_SWHID.object_id.hex()[:5] in str(e)
            assert "--\x20\n" in str(e)  # Well-formated signature!!!

        # Check that the entries have been deleted and recalling the
        # function does not re-send the e-mails
        m.reset_mock()
        swh_vault.send_notif(TEST_TYPE, TEST_SWHID)
        m.assert_not_called()


def test_send_all_emails_custom_notif(swh_vault_custom_notif):
    swh_vault = swh_vault_custom_notif
    with mock_cooking(swh_vault):
        emails = ("a@example.com", "billg@example.com", "test+42@example.org")
        for email in emails:
            swh_vault.cook(TEST_TYPE, TEST_SWHID, email=email)

    swh_vault.set_status(TEST_TYPE, TEST_SWHID, "done")

    with patch.object(swh_vault, "_smtp_send") as m:
        swh_vault.send_notif(TEST_TYPE, TEST_SWHID)

        sent_emails = {k[0][0] for k in m.call_args_list}
        assert {k["To"] for k in sent_emails} == set(emails)

        download_url = (
            "http://test.local/api/1/vault/"
            f"{TEST_TYPE.replace('_', '-')}/{str(TEST_SWHID)}/raw"
        )

        for e in sent_emails:
            assert "nobody@nowhere.local" in e["From"]
            assert TEST_TYPE in e["Subject"]
            assert TEST_SWHID.object_id.hex()[:5] in e["Subject"]
            assert TEST_TYPE in str(e)
            assert download_url in str(e)
            assert TEST_SWHID.object_id.hex()[:5] in str(e)
            assert "--\x20\n" in str(e)  # Well-formated signature!!!

        # Check that the entries have been deleted and recalling the
        # function does not re-send the e-mails
        m.reset_mock()
        swh_vault.send_notif(TEST_TYPE, TEST_SWHID)
        m.assert_not_called()


def test_send_email_error_no_smtp(swh_vault):
    reports = []
    init_sentry(SENTRY_DSN, extra_kwargs={"transport": reports.append})

    emails = ("a@example.com", "billg@example.com", "test+42@example.org")
    with mock_cooking(swh_vault):
        for email in emails:
            swh_vault.cook(TEST_TYPE, TEST_SWHID, email=email)
    swh_vault.set_status(TEST_TYPE, TEST_SWHID, "done")
    swh_vault.send_notif(TEST_TYPE, TEST_SWHID)

    assert len(reports) == 6
    for i, email in enumerate(emails):
        # first report is the logger.error
        assert reports[2 * i]["level"] == "error"
        assert reports[2 * i]["logger"] == "swh.vault.backend"
        reg = re.compile(
            f"Unable to send SMTP message 'Bundle ready: {TEST_TYPE} [0-9a-f]{{7}}' "
            f"to {email.replace('+', '[+]')}: cannot connect to server"
        )
        assert reg.match(reports[2 * i]["logentry"]["message"])
        # second is the sentry_sdk.capture_message
        assert reports[2 * i + 1]["level"] == "error"
        assert reg.match(reports[2 * i + 1]["message"])


def test_send_email_error_send_failed(swh_vault):
    reports = []
    init_sentry(SENTRY_DSN, extra_kwargs={"transport": reports.append})

    emails = ("a@example.com", "billg@example.com", "test+42@example.org")
    with mock_cooking(swh_vault):
        for email in emails:
            swh_vault.cook(TEST_TYPE, TEST_SWHID, email=email)
    swh_vault.set_status(TEST_TYPE, TEST_SWHID, "done")

    with patch("smtplib.SMTP") as MockSMTP:
        smtp = MockSMTP.return_value
        smtp.noop.return_value = [250]
        smtp.send_message.side_effect = smtplib.SMTPHeloError(404, "HELO Failed")

        swh_vault.send_notif(TEST_TYPE, TEST_SWHID)

    assert len(reports) == 4
    # first one is the captured exception
    assert reports[0]["level"] == "error"
    assert reports[0]["exception"]["values"][0]["type"] == "SMTPHeloError"

    # the following 3 ones are the sentry_sdk.capture_message() calls
    for i, email in enumerate(emails, start=1):
        assert reports[i]["level"] == "error"
        reg = re.compile(
            f"Unable to send SMTP message 'Bundle ready: {TEST_TYPE} [0-9a-f]{{7}}' "
            f"to {email.replace('+', '[+]')}: [(]404, 'HELO Failed'[)]"
        )
        assert reg.match(reports[i]["message"])


def test_available(swh_vault):
    assert not swh_vault.is_available(TEST_TYPE, TEST_SWHID)

    with mock_cooking(swh_vault):
        swh_vault.create_task(TEST_TYPE, TEST_SWHID)
    assert not swh_vault.is_available(TEST_TYPE, TEST_SWHID)

    swh_vault.cache.add(TEST_TYPE, TEST_SWHID, b"content")
    assert not swh_vault.is_available(TEST_TYPE, TEST_SWHID)

    swh_vault.set_status(TEST_TYPE, TEST_SWHID, "done")
    assert swh_vault.is_available(TEST_TYPE, TEST_SWHID)


def test_fetch(swh_vault):
    assert swh_vault.fetch(TEST_TYPE, TEST_SWHID, raise_notfound=False) is None

    with pytest.raises(
        NotFoundExc, match=f"{TEST_TYPE} {TEST_SWHID} is not available."
    ):
        swh_vault.fetch(TEST_TYPE, TEST_SWHID)

    swhid, content = fake_cook(swh_vault, TEST_TYPE, b"content")

    info = swh_vault.progress(TEST_TYPE, swhid)
    access_ts_before = info["ts_last_access"]

    assert swh_vault.fetch(TEST_TYPE, swhid) == b"content"

    info = swh_vault.progress(TEST_TYPE, swhid)
    access_ts_after = info["ts_last_access"]

    assertTimestampAlmostNow(access_ts_after)
    assert access_ts_before < access_ts_after


def test_cache_expire_oldest(swh_vault):
    r = range(1, 10)
    inserted = {}
    for i in r:
        sticky = i == 5
        content = b"content%s" % str(i).encode()
        swhid, content = fake_cook(swh_vault, TEST_TYPE, content, sticky)
        inserted[i] = (swhid, content)

    swh_vault.update_access_ts(TEST_TYPE, inserted[2][0])
    swh_vault.update_access_ts(TEST_TYPE, inserted[3][0])
    swh_vault.cache_expire_oldest(n=4)

    should_be_still_here = {2, 3, 5, 8, 9}
    for i in r:
        assert swh_vault.is_available(TEST_TYPE, inserted[i][0]) == (
            i in should_be_still_here
        )


def test_cache_expire_until(swh_vault):
    r = range(1, 10)
    inserted = {}
    for i in r:
        sticky = i == 5
        content = b"content%s" % str(i).encode()
        swhid, content = fake_cook(swh_vault, TEST_TYPE, content, sticky)
        inserted[i] = (swhid, content)

        if i == 7:
            cutoff_date = datetime.datetime.now()

    swh_vault.update_access_ts(TEST_TYPE, inserted[2][0])
    swh_vault.update_access_ts(TEST_TYPE, inserted[3][0])
    swh_vault.cache_expire_until(date=cutoff_date)

    should_be_still_here = {2, 3, 5, 8, 9}
    for i in r:
        assert swh_vault.is_available(TEST_TYPE, inserted[i][0]) == (
            i in should_be_still_here
        )


def test_fail_cook_simple(swh_vault):
    fail_cook(swh_vault, TEST_TYPE, TEST_SWHID, "error42")
    assert not swh_vault.is_available(TEST_TYPE, TEST_SWHID)
    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    assert info["progress_msg"] == "error42"


def test_send_failure_email(swh_vault):
    with mock_cooking(swh_vault):
        swh_vault.cook(TEST_TYPE, TEST_SWHID, email="a@example.com")

    swh_vault.set_status(TEST_TYPE, TEST_SWHID, "failed")
    swh_vault.set_progress(TEST_TYPE, TEST_SWHID, "test error")

    with patch.object(swh_vault, "_smtp_send") as m:
        swh_vault.send_notif(TEST_TYPE, TEST_SWHID)

        e = [k[0][0] for k in m.call_args_list][0]
        assert e["To"] == "a@example.com"

        assert "bot@softwareheritage.org" in e["From"]
        assert TEST_TYPE in e["Subject"]
        assert TEST_SWHID.object_id.hex()[:5] in e["Subject"]
        assert "fail" in e["Subject"]
        assert TEST_TYPE in str(e)
        assert TEST_SWHID.object_id.hex()[:5] in str(e)
        assert "test error" in str(e)
        assert "--\x20\n" in str(e)  # Well-formated signature


def test_retry_failed_bundle(swh_vault):
    fail_cook(swh_vault, TEST_TYPE, TEST_SWHID, "error42")
    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    assert info["task_status"] == "failed"
    with mock_cooking(swh_vault):
        swh_vault.cook(TEST_TYPE, TEST_SWHID)
    info = swh_vault.progress(TEST_TYPE, TEST_SWHID)
    assert info["task_status"] == "new"


def test_download_url_cache_pathslicing_backend(swh_vault):
    swhid, content = fake_cook(swh_vault, TEST_TYPE, b"content")
    # download URL feature is not available with pathslicing backend for vault cache
    assert swh_vault.download_url(TEST_TYPE, swhid) is None


@pytest.fixture
def swh_vault_config_http_cache(swh_vault_config, httpserver):
    swh_vault_config["cache"] = {
        "cls": "http",
        "url": httpserver.url_for("/"),
        "compression": "none",
    }
    return swh_vault_config


@pytest.fixture
def swh_vault_http_cache(swh_vault_config_http_cache):
    from swh.vault import get_vault

    return get_vault("postgresql", **swh_vault_config_http_cache)


def test_download_url_cache_http_backend(swh_vault_http_cache, mocker, httpserver):
    unknown_swhid = Content.from_data(b"foo").swhid()
    with pytest.raises(
        NotFoundExc, match=f"{TEST_TYPE} {unknown_swhid} is not available."
    ):
        swh_vault_http_cache.download_url(TEST_TYPE, unknown_swhid)

    mocker.patch.object(
        swh_vault_http_cache, "progress", return_value={"task_status": "done"}
    )
    content = b"content"
    swhid = Content.from_data(content).swhid()
    objid = swh_vault_http_cache.cache._get_internal_id(TEST_TYPE, swhid)["sha1"]
    httpserver.expect_request(f"/{objid.hex()}").respond_with_data(content)
    download_url = swh_vault_http_cache.download_url(TEST_TYPE, swhid)
    assert download_url is not None
    assert requests.get(download_url).content == content


def test_cook_if_status_done_but_bundle_not_in_cache(swh_vault, mocker):
    with mock_cooking(swh_vault):
        swh_vault.cook(TEST_TYPE, TEST_SWHID, email="a@example.com")
    swh_vault.cache.add(TEST_TYPE, TEST_SWHID, b"content")
    swh_vault.set_status(TEST_TYPE, TEST_SWHID, "done")

    create_task = mocker.spy(swh_vault, "create_task")

    with mock_cooking(swh_vault):
        swh_vault.cook(TEST_TYPE, TEST_SWHID, email="a@example.com")

    create_task.assert_not_called()

    swh_vault.cache.delete(TEST_TYPE, TEST_SWHID)

    with mock_cooking(swh_vault):
        swh_vault.cook(TEST_TYPE, TEST_SWHID, email="a@example.com")

    create_task.assert_called_once_with(TEST_TYPE, TEST_SWHID, False)


def test_registered_backends():

    entry_points = get_entry_points(group="swh.vault.classes")
    assert {ep.name for ep in entry_points} == {"remote", "postgresql", "memory"}
    for ep in entry_points:
        assert ep.load()



# Copyright (C) 2016-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import timedelta
from typing import Optional

from swh.model import hashutil
from swh.model.swhids import CoreSWHID
from swh.objstorage.factory import get_objstorage
from swh.objstorage.interface import ObjId


class VaultCache:
    """The Vault cache is an object storage that stores Vault bundles.

    This implementation computes sha1('<bundle_type>:<swhid>') as the
    internal identifiers used in the underlying objstorage.
    """

    def __init__(self, **objstorage):
        self.objstorage = get_objstorage(**objstorage)

    def add(self, bundle_type, swhid: CoreSWHID, content) -> None:
        sid = self._get_internal_id(bundle_type, swhid)
        self.objstorage.add(content, sid)

    def get(self, bundle_type, swhid: CoreSWHID) -> bytes:
        sid = self._get_internal_id(bundle_type, swhid)
        return self.objstorage.get(sid)

    def download_url(
        self,
        bundle_type,
        swhid: CoreSWHID,
        content_disposition: Optional[str] = None,
        expiry: Optional[timedelta] = None,
    ) -> Optional[str]:
        sid = self._get_internal_id(bundle_type, swhid)
        return self.objstorage.download_url(sid, content_disposition, expiry)

    def delete(self, bundle_type, swhid: CoreSWHID):
        sid = self._get_internal_id(bundle_type, swhid)
        return self.objstorage.delete(sid)

    def is_cached(self, bundle_type, swhid: CoreSWHID) -> bool:
        sid = self._get_internal_id(bundle_type, swhid)
        return sid in self.objstorage

    def _get_internal_id(self, bundle_type, swhid: CoreSWHID) -> ObjId:
        key = "{}:{}".format(bundle_type, swhid).encode()
        return hashutil.MultiHash.from_data(key).digest()



# Copyright (C) 2017-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.model.swhids import CoreSWHID

TEST_TYPE_1 = "revision_gitfast"
TEST_TYPE_2 = "directory"

TEST_SWHID_1 = CoreSWHID.from_string(
    "swh:1:rev:4a4b9771542143cf070386f86b4b92d42966bdbc"
)
TEST_SWHID_2 = CoreSWHID.from_string(
    "swh:1:dir:17a3e48bce37be5226490e750202ad3a9a1a3fe9"
)

TEST_CONTENT_1 = b"test content 1"
TEST_CONTENT_2 = b"test content 2"


# Let's try to avoid replicating edge-cases already tested in
# swh-objstorage, and instead focus on testing behaviors specific to the
# Vault cache here.


def test_internal_id(swh_vault):
    sid = swh_vault.cache._get_internal_id(TEST_TYPE_1, TEST_SWHID_1)
    assert sid["sha1"].hex() == "ec2a99d6b21a68648a9d0c99c5d7c35f69268564"


def test_simple_add_get(swh_vault):
    swh_vault.cache.add(TEST_TYPE_1, TEST_SWHID_1, TEST_CONTENT_1)
    assert swh_vault.cache.get(TEST_TYPE_1, TEST_SWHID_1) == TEST_CONTENT_1
    assert swh_vault.cache.is_cached(TEST_TYPE_1, TEST_SWHID_1)


def test_different_type_same_id(swh_vault):
    swh_vault.cache.add(TEST_TYPE_1, TEST_SWHID_1, TEST_CONTENT_1)
    swh_vault.cache.add(TEST_TYPE_2, TEST_SWHID_1, TEST_CONTENT_2)
    assert swh_vault.cache.get(TEST_TYPE_1, TEST_SWHID_1) == TEST_CONTENT_1
    assert swh_vault.cache.get(TEST_TYPE_2, TEST_SWHID_1) == TEST_CONTENT_2
    assert swh_vault.cache.is_cached(TEST_TYPE_1, TEST_SWHID_1)
    assert swh_vault.cache.is_cached(TEST_TYPE_2, TEST_SWHID_1)


def test_different_type_same_content(swh_vault):
    swh_vault.cache.add(TEST_TYPE_1, TEST_SWHID_1, TEST_CONTENT_1)
    swh_vault.cache.add(TEST_TYPE_2, TEST_SWHID_1, TEST_CONTENT_1)
    assert swh_vault.cache.get(TEST_TYPE_1, TEST_SWHID_1) == TEST_CONTENT_1
    assert swh_vault.cache.get(TEST_TYPE_2, TEST_SWHID_1) == TEST_CONTENT_1
    assert swh_vault.cache.is_cached(TEST_TYPE_1, TEST_SWHID_1)
    assert swh_vault.cache.is_cached(TEST_TYPE_2, TEST_SWHID_1)


def test_different_id_same_type(swh_vault):
    swh_vault.cache.add(TEST_TYPE_1, TEST_SWHID_1, TEST_CONTENT_1)
    swh_vault.cache.add(TEST_TYPE_1, TEST_SWHID_2, TEST_CONTENT_2)
    assert swh_vault.cache.get(TEST_TYPE_1, TEST_SWHID_1) == TEST_CONTENT_1
    assert swh_vault.cache.get(TEST_TYPE_1, TEST_SWHID_2) == TEST_CONTENT_2
    assert swh_vault.cache.is_cached(TEST_TYPE_1, TEST_SWHID_1)
    assert swh_vault.cache.is_cached(TEST_TYPE_1, TEST_SWHID_2)



# Copyright (C) 2018-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU Affero General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict

if TYPE_CHECKING:
    from .interface import VaultInterface

logger = logging.getLogger(__name__)


def get_vault(cls: str, **kwargs) -> "VaultInterface":
    """
    Get a vault object of class `vault_class` with arguments
    `vault_args`.

    Args:
        cls: vault's class
        kwargs: arguments to pass to the class' constructor

    Returns:
        an instance of VaultBackend

    Raises:
        ValueError if passed an unknown storage class.

    """
    from swh.core.config import get_swh_backend_module

    _, Vault = get_swh_backend_module("vault", cls)
    assert Vault is not None
    return Vault(**kwargs)


default_cfg = {
    "default_interval": "1 day",
    "min_interval": "1 day",
    "max_interval": "1 day",
    "backoff_factor": 1,
    "max_queue_length": 10000,
}


def register_tasks() -> Dict[str, Any]:
    return {
        "task_modules": [f"{__name__}.cooking_tasks"],
        "task_types": {
            "vault-cook-bundle": default_cfg,
            "vault-batch-cook-bundle": default_cfg,
        },
    }



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
This module contains additional tests for the bare cooker.
Generic cooker tests (eg. without swh-graph) in test_cookers.py also
run on the bare cooker.
"""

import datetime
import enum
from functools import partial
import io
import os
import subprocess
import tarfile
import tempfile
import unittest.mock

import attr
import pytest
from pytest import param
from pytest_postgresql import factories

from swh.core.db.db_utils import initialize_database_for_module
from swh.model.from_disk import DentryPerms
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    ObjectType,
    Person,
    Release,
    Revision,
    RevisionType,
    Snapshot,
    SnapshotBranch,
    SnapshotTargetType,
    TimestampWithTimezone,
)
from swh.storage import get_storage
from swh.storage.postgresql.storage import Storage
from swh.vault.cookers.git_bare import GitBareCooker
from swh.vault.in_memory_backend import InMemoryVaultBackend

storage_postgresql_proc = factories.postgresql_proc(
    load=[partial(initialize_database_for_module, "storage", Storage.current_version)],
)

storage_postgresql = factories.postgresql("storage_postgresql_proc")


@pytest.fixture
def swh_storage(storage_postgresql):
    return get_storage(
        "postgresql", db=storage_postgresql.info.dsn, objstorage={"cls": "memory"}
    )


class RootObjects(enum.Enum):
    REVISION = enum.auto()
    SNAPSHOT = enum.auto()
    RELEASE = enum.auto()
    WEIRD_RELEASE = enum.auto()  # has a : in the name + points to another release


@pytest.mark.graph
@pytest.mark.parametrize(
    "root_object,up_to_date_graph,tag,weird_branches",
    [
        param(
            RootObjects.REVISION,
            False,
            False,
            False,
            id="rev, outdated graph, no tag/tree/blob",
        ),
        param(
            RootObjects.REVISION,
            True,
            False,
            False,
            id="rev, updated graph, no tag/tree/blob",
        ),
        param(
            RootObjects.RELEASE,
            False,
            False,
            False,
            id="rel, outdated graph, no tag/tree/blob",
        ),
        param(
            RootObjects.RELEASE,
            True,
            False,
            False,
            id="rel, updated graph, no tag/tree/blob",
        ),
        param(
            RootObjects.WEIRD_RELEASE,
            True,
            False,
            False,
            id="weird rel, updated graph, no tag/tree/blob",
        ),
        param(
            RootObjects.SNAPSHOT,
            False,
            False,
            False,
            id="snp, outdated graph, no tag/tree/blob",
        ),
        param(
            RootObjects.SNAPSHOT,
            True,
            False,
            False,
            id="snp, updated graph, no tag/tree/blob",
        ),
        param(
            RootObjects.SNAPSHOT,
            False,
            True,
            False,
            id="snp, outdated graph, w/ tag, no tree/blob",
        ),
        param(
            RootObjects.SNAPSHOT,
            True,
            True,
            False,
            id="snp, updated graph, w/ tag, no tree/blob",
        ),
        param(
            RootObjects.SNAPSHOT,
            False,
            True,
            True,
            id="snp, outdated graph, w/ tag, tree, and blob",
        ),
        param(
            RootObjects.SNAPSHOT,
            True,
            True,
            True,
            id="snp, updated graph, w/ tag, tree, and blob",
        ),
    ],
)
def test_graph_revisions(
    swh_storage, up_to_date_graph, root_object, tag, weird_branches
):
    r"""
    Build objects::

                                     snp
                                    /|||\
                                   / ||| \
                        rel2 <----°  /|\  \----> rel4
                         |          / | \         |
                         v         /  v  \        v
          rev1  <------ rev2 <----°  dir4 \      rel3
           |             |            |    \      |
           v             v            v     \     |
          dir1          dir2         dir3   |     |
           |           /   |          |     |     |
           v          /    v          v     v     v
          cnt1  <----°    cnt2       cnt3  cnt4  cnt5

    If up_to_date_graph is true, then swh-graph contains all objects.
    Else, cnt4, cnt5, dir4, rev2, rel2, rel3, and snp are missing from the graph.

    If tag is False, rel2 is excluded.

    If weird_branches is False, dir4, cnt4, rel3, rel4, and cnt5 are excluded.
    """
    from swh.graph.http_naive_client import NaiveClient as GraphClient

    # Create objects:

    date = TimestampWithTimezone.from_datetime(
        datetime.datetime(2021, 5, 7, 8, 43, 59, tzinfo=datetime.timezone.utc)
    )
    author = Person.from_fullname(b"Foo <foo@example.org>")
    cnt1 = Content.from_data(b"correct")
    cnt2 = Content.from_data(b"horse")
    cnt3 = Content.from_data(b"battery")
    cnt4 = Content.from_data(b"staple")
    cnt5 = Content.from_data(b"Tr0ub4dor&3")
    dir1 = Directory(
        entries=(
            DirectoryEntry(
                name=b"file1",
                type="file",
                perms=DentryPerms.content,
                target=cnt1.sha1_git,
            ),
        )
    )
    dir2 = Directory(
        entries=(
            DirectoryEntry(
                name=b"file1",
                type="file",
                perms=DentryPerms.content,
                target=cnt1.sha1_git,
            ),
            DirectoryEntry(
                name=b"file2",
                type="file",
                perms=DentryPerms.content,
                target=cnt2.sha1_git,
            ),
        )
    )
    dir3 = Directory(
        entries=(
            DirectoryEntry(
                name=b"file3",
                type="file",
                perms=DentryPerms.content,
                target=cnt3.sha1_git,
            ),
        )
    )
    dir4 = Directory(
        entries=(
            DirectoryEntry(
                name=b"directory3",
                type="dir",
                perms=DentryPerms.directory,
                target=dir3.id,
            ),
        )
    )
    rev1 = Revision(
        message=b"msg1",
        date=date,
        committer_date=date,
        author=author,
        committer=author,
        directory=dir1.id,
        type=RevisionType.GIT,
        synthetic=True,
    )
    rev2 = Revision(
        message=b"msg2",
        date=date,
        committer_date=date,
        author=author,
        committer=author,
        directory=dir2.id,
        parents=(rev1.id,),
        type=RevisionType.GIT,
        synthetic=True,
    )

    rel2 = Release(
        name=b"1.0.0",
        message=b"tag2",
        target_type=ObjectType.REVISION,
        target=rev2.id,
        synthetic=True,
    )
    rel3 = Release(
        name=b"1.0.0-blob",
        message=b"tagged-blob",
        target_type=ObjectType.CONTENT,
        target=cnt5.sha1_git,
        synthetic=True,
    )
    rel4 = Release(
        name=b"1.0.0-weird",
        message=b"weird release",
        target_type=ObjectType.RELEASE,
        target=rel3.id,
        synthetic=True,
    )
    rel5 = Release(
        name=b"1.0.0:weirdname",
        message=b"weird release",
        target_type=ObjectType.RELEASE,
        target=rel2.id,
        synthetic=True,
    )

    # Create snapshot:

    branches = {
        b"refs/heads/master": SnapshotBranch(
            target=rev2.id, target_type=SnapshotTargetType.REVISION
        ),
    }
    if tag:
        branches[b"refs/tags/1.0.0"] = SnapshotBranch(
            target=rel2.id, target_type=SnapshotTargetType.RELEASE
        )
    if weird_branches:
        branches[b"refs/heads/tree-ref"] = SnapshotBranch(
            target=dir4.id, target_type=SnapshotTargetType.DIRECTORY
        )
        branches[b"refs/heads/blob-ref"] = SnapshotBranch(
            target=cnt4.sha1_git, target_type=SnapshotTargetType.CONTENT
        )
        branches[b"refs/tags/1.0.0-weird"] = SnapshotBranch(
            target=rel4.id, target_type=SnapshotTargetType.RELEASE
        )
    snp = Snapshot(branches=branches)

    # "Fill" swh-graph

    if up_to_date_graph:
        nodes = [cnt1, cnt2, dir1, dir2, rev1, rev2, snp]
        edges = [
            (dir1, cnt1),
            (dir2, cnt1),
            (dir2, cnt2),
            (rev1, dir1),
            (rev2, dir2),
            (rev2, rev1),
            (snp, rev2),
        ]
        if tag:
            nodes.append(rel2)
            edges.append((rel2, rev2))
            edges.append((snp, rel2))
        if weird_branches:
            nodes.extend([cnt3, cnt4, cnt5, dir3, dir4, rel3, rel4, rel5])
            edges.extend(
                [
                    (dir3, cnt3),
                    (dir4, dir3),
                    (snp, dir4),
                    (snp, cnt4),
                    (snp, rel4),
                    (rel4, rel3),
                    (rel3, cnt5),
                    (rel5, rev2),
                ]
            )
    else:
        nodes = [cnt1, cnt2, cnt3, dir1, dir2, dir3, rev1]
        edges = [
            (dir1, cnt1),
            (dir2, cnt1),
            (dir2, cnt2),
            (dir3, cnt3),
            (rev1, dir1),
        ]
        if tag:
            nodes.append(rel2)
        if weird_branches:
            nodes.extend([cnt3, dir3])
            edges.extend([(dir3, cnt3)])

    nodes = [str(n.swhid()) for n in nodes]
    edges = [(str(s.swhid()), str(d.swhid())) for (s, d) in edges]

    # Add all objects to storage
    swh_storage.content_add([cnt1, cnt2, cnt3, cnt4, cnt5])
    swh_storage.directory_add([dir1, dir2, dir3, dir4])
    swh_storage.revision_add([rev1, rev2])
    swh_storage.release_add([rel2, rel3, rel4, rel5])
    swh_storage.snapshot_add([snp])

    # Add spy on swh_storage, to make sure revision_log is not called
    # (the graph must be used instead)
    swh_storage = unittest.mock.MagicMock(wraps=swh_storage)

    # Add all objects to graph
    swh_graph = unittest.mock.Mock(wraps=GraphClient(nodes=nodes, edges=edges))

    # Cook
    backend = InMemoryVaultBackend()
    cooked_swhid = {
        RootObjects.SNAPSHOT: snp.swhid(),
        RootObjects.REVISION: rev2.swhid(),
        RootObjects.RELEASE: rel2.swhid(),
        RootObjects.WEIRD_RELEASE: rel5.swhid(),
    }[root_object]
    cooker = GitBareCooker(
        cooked_swhid,
        backend=backend,
        storage=swh_storage,
        graph=swh_graph,
    )

    if weird_branches:
        # git-fsck now rejects refs pointing to trees and blobs,
        # but some old git repos have them.
        cooker.use_fsck = False

    cooker.cook()

    # Get bundle
    bundle = backend.fetch("git_bare", cooked_swhid)

    # Extract bundle and make sure both revisions are in it
    with tempfile.TemporaryDirectory("swh-vault-test-bare") as tempdir:
        with tarfile.open(fileobj=io.BytesIO(bundle)) as tf:
            tf.extractall(tempdir)

        if root_object != RootObjects.WEIRD_RELEASE:
            # check master ref exists in repository
            master_ref_path = os.path.join(
                tempdir, f"{cooked_swhid}.git/refs/heads/master"
            )
            assert os.path.exists(master_ref_path)
            with open(master_ref_path, "r") as master_ref:
                assert master_ref.read() == branches[b"refs/heads/master"].target.hex()

        if root_object in (RootObjects.SNAPSHOT, RootObjects.REVISION):
            log_head = "master"
        elif root_object == RootObjects.RELEASE:
            log_head = "1.0.0"
        elif root_object == RootObjects.WEIRD_RELEASE:
            log_head = "release"
        else:
            assert False, root_object

        output = subprocess.check_output(
            [
                "git",
                "-C",
                f"{tempdir}/{cooked_swhid}.git",
                "log",
                "--format=oneline",
                "--decorate=",
                log_head,
            ]
        )

        assert output.decode() == f"{rev2.id.hex()} msg2\n{rev1.id.hex()} msg1\n"

    # Make sure the graph was used instead of swh_storage.revision_log
    if root_object == RootObjects.SNAPSHOT:
        if up_to_date_graph:
            # The graph has everything, so the first call succeeds and returns
            # all objects transitively pointed by the snapshot
            swh_graph.visit_nodes.assert_has_calls(
                [
                    unittest.mock.call(str(snp.swhid()), edges="snp:*,rel:*,rev:rev"),
                ]
            )
        else:
            # The graph does not have everything, so the first call returns nothing.
            # However, the second call (on the top rev) succeeds and returns
            # all objects but the rev and the rel
            swh_graph.visit_nodes.assert_has_calls(
                [
                    unittest.mock.call(str(snp.swhid()), edges="snp:*,rel:*,rev:rev"),
                    unittest.mock.call(str(rev2.swhid()), edges="rev:rev"),
                ]
            )
    elif root_object in (
        RootObjects.REVISION,
        RootObjects.RELEASE,
        RootObjects.WEIRD_RELEASE,
    ):
        swh_graph.visit_nodes.assert_has_calls(
            [unittest.mock.call(str(rev2.swhid()), edges="rev:rev")]
        )
    else:
        assert False, root_object

    if up_to_date_graph:
        swh_storage.revision_log.assert_not_called()
        swh_storage.revision_shortlog.assert_not_called()
    else:
        swh_storage.revision_log.assert_called()


@pytest.mark.parametrize(
    "mismatch_on", ["content", "directory", "revision1", "revision2", "none"]
)
def test_checksum_mismatch(swh_storage, mismatch_on):
    date = TimestampWithTimezone.from_datetime(
        datetime.datetime(2021, 5, 7, 8, 43, 59, tzinfo=datetime.timezone.utc)
    )
    author = Person.from_fullname(b"Foo <foo@example.org>")

    wrong_hash = b"\x12\x34" * 10

    cnt1 = Content.from_data(b"Tr0ub4dor&3")
    if mismatch_on == "content":
        cnt1 = attr.evolve(cnt1, sha1_git=wrong_hash)

    dir1 = Directory(
        entries=(
            DirectoryEntry(
                name=b"file1",
                type="file",
                perms=DentryPerms.content,
                target=cnt1.sha1_git,
            ),
        )
    )

    if mismatch_on == "directory":
        dir1 = attr.evolve(dir1, id=wrong_hash)

    rev1 = Revision(
        message=b"msg1",
        date=date,
        committer_date=date,
        author=author,
        committer=author,
        directory=dir1.id,
        type=RevisionType.GIT,
        synthetic=True,
    )

    if mismatch_on == "revision1":
        rev1 = attr.evolve(rev1, id=wrong_hash)

    rev2 = Revision(
        message=b"msg2",
        date=date,
        committer_date=date,
        author=author,
        committer=author,
        directory=dir1.id,
        parents=(rev1.id,),
        type=RevisionType.GIT,
        synthetic=True,
    )

    if mismatch_on == "revision2":
        rev2 = attr.evolve(rev2, id=wrong_hash)

    cooked_swhid = rev2.swhid()

    swh_storage.content_add([cnt1])
    swh_storage.directory_add([dir1])
    swh_storage.revision_add([rev1, rev2])

    backend = InMemoryVaultBackend()
    cooker = GitBareCooker(
        cooked_swhid,
        backend=backend,
        storage=swh_storage,
        graph=None,
    )

    cooker.cook()

    # Get bundle
    bundle = backend.fetch("git_bare", cooked_swhid)

    # Extract bundle and make sure both revisions are in it
    with tempfile.TemporaryDirectory("swh-vault-test-bare") as tempdir:
        with tarfile.open(fileobj=io.BytesIO(bundle)) as tf:
            tf.extractall(tempdir)

        if mismatch_on != "revision2":
            # git-log fails if the head revision is corrupted
            # TODO: we need to find a way to make this somewhat usable
            output = subprocess.check_output(
                [
                    "git",
                    "-C",
                    f"{tempdir}/{cooked_swhid}.git",
                    "log",
                    "--format=oneline",
                    "--decorate=",
                ]
            )

            assert output.decode() == f"{rev2.id.hex()} msg2\n{rev1.id.hex()} msg1\n"



# Copyright (C) 2020-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import os
from typing import Any, Dict

import pytest
import yaml

from swh.core.api.serializers import json_dumps, msgpack_dumps, msgpack_loads
from swh.vault.api.serializers import ENCODERS
import swh.vault.api.server
from swh.vault.api.server import app, check_config, get_vault, make_app_from_configfile
from swh.vault.tests.test_backend import TEST_SWHID


@pytest.fixture
def swh_vault_server_config(swh_vault_config: Dict[str, Any]) -> Dict[str, Any]:
    """Returns a vault server configuration, with ``storage``, ``scheduler`` and
    ``cache`` set at the toplevel"""
    return {
        "vault": {"cls": "postgresql", "db": swh_vault_config["db"]},
        "client_max_size": 1024**3,
        **{k: v for k, v in swh_vault_config.items() if k != "db"},
    }


@pytest.fixture
def swh_vault_server_config_file(swh_vault_server_config, monkeypatch, tmp_path):
    """Creates a vault server configuration file and sets it into SWH_CONFIG_FILENAME"""
    conf_path = os.path.join(str(tmp_path), "vault-server.yml")
    with open(conf_path, "w") as f:
        f.write(yaml.dump(swh_vault_server_config))
    monkeypatch.setenv("SWH_CONFIG_FILENAME", conf_path)
    return conf_path


def test_make_app_from_file_missing(monkeypatch):
    monkeypatch.delenv("SWH_CONFIG_FILENAME", raising=False)
    with pytest.raises(ValueError, match="Missing configuration path."):
        make_app_from_configfile()


def test_make_app_from_file_does_not_exist(tmp_path, monkeypatch):
    monkeypatch.delenv("SWH_CONFIG_FILENAME", raising=False)
    conf_path = os.path.join(str(tmp_path), "vault-server.yml")
    assert os.path.exists(conf_path) is False

    with pytest.raises(
        ValueError, match=f"Configuration path {conf_path} should exist."
    ):
        make_app_from_configfile(conf_path)


def test_make_app_from_env_variable(swh_vault_server_config_file):
    """Server initialization happens through env variable when no path is provided"""
    app = make_app_from_configfile()
    assert app is not None
    assert get_vault() is not None

    # Cleanup app
    del app.config["vault"]
    swh.vault.api.server.vault = None


def test_make_app_from_file(swh_vault_server_config, tmp_path, monkeypatch):
    """Server initialization happens through path if provided"""
    monkeypatch.delenv("SWH_CONFIG_FILENAME", raising=False)
    conf_path = os.path.join(str(tmp_path), "vault-server.yml")
    with open(conf_path, "w") as f:
        f.write(yaml.dump(swh_vault_server_config))

    app = make_app_from_configfile(conf_path)
    assert app is not None
    assert get_vault() is not None

    # Cleanup app
    del app.config["vault"]
    swh.vault.api.server.vault = None


@pytest.fixture
def vault_app(swh_vault_server_config_file):
    yield make_app_from_configfile()

    # Cleanup app
    del app.config["vault"]
    swh.vault.api.server.vault = None


@pytest.fixture
def cli(vault_app):
    cli = vault_app.test_client()
    return cli


def test_client_index(cli):
    resp = cli.get("/")
    assert resp.status == "200 OK"


def test_client_cook_notfound(cli):
    resp = cli.post(
        "/cook",
        data=json_dumps(
            {"bundle_type": "flat", "swhid": TEST_SWHID}, extra_encoders=ENCODERS
        ),
        headers=[("Content-Type", "application/json")],
    )
    assert resp.status == "400 BAD REQUEST"
    content = msgpack_loads(resp.data)
    assert content["type"] == "NotFoundExc"
    assert content["args"] == [f"flat {TEST_SWHID} was not found."]


def test_client_progress_notfound(cli):
    resp = cli.post(
        "/progress",
        data=json_dumps(
            {"bundle_type": "flat", "swhid": TEST_SWHID}, extra_encoders=ENCODERS
        ),
        headers=[("Content-Type", "application/json")],
    )
    assert resp.status == "400 BAD REQUEST"
    content = msgpack_loads(resp.data)
    assert content["type"] == "NotFoundExc"
    assert content["args"] == [f"flat {TEST_SWHID} was not found."]


def test_client_batch_cook_invalid_type(cli):
    resp = cli.post(
        "/batch_cook",
        data=msgpack_dumps({"batch": [("foobar", [])]}),
        headers={"Content-Type": "application/x-msgpack"},
    )
    assert resp.status == "400 BAD REQUEST"
    content = msgpack_loads(resp.data)
    assert content["type"] == "NotFoundExc"
    assert content["args"] == ["foobar is an unknown type."]


def test_client_batch_progress_notfound(cli):
    resp = cli.post(
        "/batch_progress",
        data=msgpack_dumps({"batch_id": 1}),
        headers={"Content-Type": "application/x-msgpack"},
    )
    assert resp.status == "400 BAD REQUEST"
    content = msgpack_loads(resp.data)
    assert content["type"] == "NotFoundExc"
    assert content["args"] == ["Batch 1 does not exist."]


def test_check_config_missing_vault_configuration() -> None:
    """Irrelevant configuration file path raises"""
    with pytest.raises(ValueError, match="missing 'vault' configuration"):
        check_config({})


def test_check_config_not_local() -> None:
    """Wrong configuration raises"""
    expected_error = (
        "The vault backend of a vault server cannot be a 'remote' configuration"
    )
    with pytest.raises(EnvironmentError, match=expected_error):
        check_config({"vault": {"cls": "remote"}})


def test_check_config_ok(swh_vault_server_config) -> None:
    """Check that the default config is accepted"""
    assert swh_vault_server_config["vault"]["cls"] == "postgresql"
    assert check_config(swh_vault_server_config) is not None


@pytest.mark.parametrize("missing_key", ["storage", "cache", "scheduler"])
def test_check_config_missing_key(missing_key, swh_vault_server_config) -> None:
    """Check that configs with a missing key get rejected"""
    config_ok = swh_vault_server_config
    config_ko = copy.deepcopy(config_ok)
    config_ko["vault"].pop(missing_key, None)
    config_ko.pop(missing_key, None)

    expected_error = f"invalid configuration: missing {missing_key} config entry"
    with pytest.raises(ValueError, match=expected_error):
        check_config(config_ko)



# Copyright (C) 2020-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import partial
import os
from typing import Any, Dict

import pytest
from pytest_postgresql import factories

from swh.core.db.db_utils import initialize_database_for_module
from swh.vault import get_vault
from swh.vault.backend import VaultBackend

os.environ["LC_ALL"] = "C.UTF-8"

# needed for directory tests on git-cloned repositories
# 022 is usually the default value, but some environments (eg. Debian builds) have
# a different one.
os.umask(0o022)


vault_postgresql_proc = factories.postgresql_proc(
    load=[
        partial(initialize_database_for_module, "vault", VaultBackend.current_version)
    ],
)

postgres_vault = factories.postgresql("vault_postgresql_proc")


def pytest_collection_modifyitems(items):
    """Skip tests using httpserver fixture if pytest-httpserver is
    not available (debian < 12 for instance)"""
    try:
        from pytest_httpserver import HTTPServer  # noqa
    except ImportError:
        pytest_httpserver_available = False
    else:
        pytest_httpserver_available = True
    for item in items:
        try:
            fixtures = item.fixturenames
            if "httpserver" in fixtures and not pytest_httpserver_available:
                item.add_marker(
                    pytest.mark.skip(reason="pytest-httpserver not installed")
                )
        except Exception:
            pass


@pytest.fixture
def swh_vault_config(postgres_vault, tmp_path) -> Dict[str, Any]:
    tmp_path = str(tmp_path)
    return {
        "db": postgres_vault.info.dsn,
        "storage": {
            "cls": "memory",
        },
        "cache": {
            "cls": "pathslicing",
            "root": tmp_path,
            "slicing": "0:1/1:5",
            "allow_delete": True,
        },
        "scheduler": {
            "cls": "remote",
            "url": "http://swh-scheduler:5008",
        },
    }


@pytest.fixture
def swh_vault(swh_vault_config):
    return get_vault("postgresql", **swh_vault_config)


@pytest.fixture
def swh_vault_custom_notif(swh_vault_config):
    notif_cfg = {
        "from": "Someone from somewhere <nobody@nowhere.local>",
        "api_url": "http://test.local/api/1",
    }
    return get_vault("postgresql", notification=notif_cfg, **swh_vault_config)


@pytest.fixture
def swh_storage(swh_vault):
    return swh_vault.storage



# Copyright (C) 2016-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import io
import logging
import traceback
from typing import ClassVar, Optional, Set

from psycopg2.extensions import QueryCanceledError
import sentry_sdk

import swh.model.swhids
from swh.model.swhids import CoreSWHID, ObjectType
from swh.objstorage.interface import ObjStorageInterface
from swh.storage.interface import StorageInterface

MAX_BUNDLE_SIZE = 2**29  # 512 MiB
DEFAULT_CONFIG_PATH = "vault/cooker"
DEFAULT_CONFIG = {
    "max_bundle_size": ("int", MAX_BUNDLE_SIZE),
}


class PolicyError(Exception):
    """Raised when the bundle violates the cooking policy."""

    pass


class BundleTooLargeError(PolicyError):
    """Raised when the bundle is too large to be cooked."""

    pass


class BytesIOBundleSizeLimit(io.BytesIO):
    def __init__(self, *args, size_limit=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.size_limit = size_limit

    def write(self, chunk):
        if (
            self.size_limit is not None
            and self.getbuffer().nbytes + len(chunk) > self.size_limit
        ):
            raise BundleTooLargeError(
                "The requested bundle exceeds the maximum allowed "
                "size of {} bytes.".format(self.size_limit)
            )
        return super().write(chunk)


class BaseVaultCooker(metaclass=abc.ABCMeta):
    """Abstract base class for the vault's bundle creators

    This class describes a common API for the cookers.

    To define a new cooker, inherit from this class and override:
    - CACHE_TYPE_KEY: key to use for the bundle to reference in cache
    - def cook(): cook the object into a bundle
    """

    SUPPORTED_OBJECT_TYPES: ClassVar[Set[swh.model.swhids.ObjectType]]
    BUNDLE_TYPE: ClassVar[str]

    def __init__(
        self,
        swhid: CoreSWHID,
        backend,
        storage: StorageInterface,
        graph=None,
        objstorage: Optional[ObjStorageInterface] = None,
        max_bundle_size: int = MAX_BUNDLE_SIZE,
        thread_pool_size: int = 10,
    ):
        """Initialize the cooker.

        The type of the object represented by the id depends on the
        concrete class. Very likely, each type of bundle will have its
        own cooker class.

        Args:
            swhid: id of the object to be cooked into a bundle.
            backend: the vault backend (swh.vault.backend.VaultBackend).
        """
        self.check_object_type(swhid.object_type)
        self.swhid = swhid
        self.obj_id = swhid.object_id
        self.backend = backend
        self.storage = storage
        self.objstorage = objstorage
        self.graph = graph
        self.max_bundle_size = max_bundle_size
        self.thread_pool_size = thread_pool_size

    @classmethod
    def check_object_type(cls, object_type: ObjectType) -> None:
        if object_type not in cls.SUPPORTED_OBJECT_TYPES:
            raise ValueError(f"{cls.__name__} does not support {object_type} objects.")

    @abc.abstractmethod
    def check_exists(self):
        """Checks that the requested object exists and can be cooked.

        Override this in the cooker implementation.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def prepare_bundle(self):
        """Implementation of the cooker. Yields chunks of the bundle bytes.

        Override this with the cooker implementation.
        """
        raise NotImplementedError

    def cache_type_key(self) -> str:
        assert self.BUNDLE_TYPE
        return self.BUNDLE_TYPE

    def write(self, chunk):
        self.fileobj.write(chunk)

    def cook(self):
        """Cook the requested object into a bundle"""
        self.backend.set_status(self.BUNDLE_TYPE, self.swhid, "pending")
        self.backend.set_progress(self.BUNDLE_TYPE, self.swhid, "Processing...")

        self.fileobj = BytesIOBundleSizeLimit(size_limit=self.max_bundle_size)
        try:
            try:
                self.prepare_bundle()
            except QueryCanceledError:
                raise PolicyError(
                    "Timeout reached while assembling the requested bundle"
                )
            bundle = self.fileobj.getvalue()
            # TODO: use proper content streaming instead of put_bundle()
            self.backend.put_bundle(self.cache_type_key(), self.swhid, bundle)
        except PolicyError as e:
            logging.info("Bundle cooking violated policy: %s", e)
            self.backend.set_status(self.BUNDLE_TYPE, self.swhid, "failed")
            self.backend.set_progress(self.BUNDLE_TYPE, self.swhid, str(e))
        except Exception:
            self.backend.set_status(self.BUNDLE_TYPE, self.swhid, "failed")
            tb = traceback.format_exc()
            self.backend.set_progress(
                self.BUNDLE_TYPE,
                self.swhid,
                f"Internal Server Error. This incident will be reported.\n"
                f"The full error was:\n\n{tb}",
            )
            logging.exception("Bundle cooking failed.")
            sentry_sdk.capture_exception()
        else:
            self.backend.set_status(self.BUNDLE_TYPE, self.swhid, "done")
            self.backend.set_progress(self.BUNDLE_TYPE, self.swhid, None)
        finally:
            self.backend.send_notif(self.BUNDLE_TYPE, self.swhid)



# Copyright (C) 2016  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import tarfile
import tempfile

from swh.model.swhids import ObjectType
from swh.vault.cookers.base import BaseVaultCooker
from swh.vault.to_disk import DirectoryBuilder


class DirectoryCooker(BaseVaultCooker):
    """Cooker to create a directory bundle"""

    BUNDLE_TYPE = "flat"
    SUPPORTED_OBJECT_TYPES = {ObjectType.DIRECTORY}

    def check_exists(self):
        return not list(self.storage.directory_missing([self.obj_id]))

    def prepare_bundle(self):
        with tempfile.TemporaryDirectory(prefix="tmp-vault-directory-") as td:
            directory_builder = DirectoryBuilder(
                storage=self.storage,
                root=td.encode(),
                dir_id=self.obj_id,
                thread_pool_size=self.thread_pool_size,
                objstorage=self.objstorage,
            )
            directory_builder.build()
            with tarfile.open(fileobj=self.fileobj, mode="w:gz") as tar:
                tar.add(td, arcname=str(self.swhid))



# Copyright (C) 2016-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from celery import current_app as app

from swh.model.swhids import CoreSWHID
from swh.vault.cookers import get_cooker


@app.task(name=__name__ + ".SWHCookingTask")
def cook_vault_bundle(bundle_type, swhid):
    """Main task to cook a bundle."""
    get_cooker(bundle_type, CoreSWHID.from_string(swhid)).cook()


# TODO: remove once the scheduler handles priority tasks
@app.task(name=__name__ + ".SWHBatchCookingTask")
def cook_vault_bundle_batch(bundle_type, swhid):
    """Temporary task for the batch queue."""
    get_cooker(bundle_type, CoreSWHID.from_string(swhid)).cook()



# Copyright (C) 2016-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.core.api import RPCClient
from swh.vault.exc import NotFoundExc
from swh.vault.interface import VaultInterface

from .serializers import DECODERS, ENCODERS


class RemoteVaultClient(RPCClient):
    """Client to the Software Heritage vault cache."""

    backend_class = VaultInterface
    reraise_exceptions = [NotFoundExc]
    extra_type_decoders = DECODERS
    extra_type_encoders = ENCODERS



# Copyright (C) 2016-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import concurrent
import os
from typing import Any, Dict, Optional

from swh.model import hashutil
from swh.model.from_disk import DentryPerms, mode_to_perms
from swh.objstorage.interface import ObjStorageInterface, objid_from_dict
from swh.storage.interface import StorageInterface

MISSING_MESSAGE = (
    b"This content is missing from the Software Heritage archive "
    b"(or from the mirror used while retrieving it)."
)

SKIPPED_MESSAGE = (
    b"This content has not been retrieved in the "
    b"Software Heritage archive due to its size."
)

HIDDEN_MESSAGE = b"This content is hidden."


def get_filtered_file_content(
    storage: StorageInterface,
    file_data: Dict[str, Any],
    objstorage: Optional[ObjStorageInterface] = None,
) -> Dict[str, Any]:
    """Retrieve the file specified by file_data and apply filters for skipped
    and missing content.

    Args:
        storage: the storage from which to retrieve the objects
        file_data: a file entry as returned by directory_ls()

    Returns:
        The entry given in file_data with a new 'content' key that points to
        the file content in bytes.

        The contents can be replaced by a specific message to indicate that
        they could not be retrieved (either due to privacy policy or because
        their sizes were too big for us to archive it).

    """
    status = file_data["status"]
    if status == "visible":
        hashes = objid_from_dict(file_data)
        data: Optional[bytes]
        if objstorage is not None:
            data = objstorage.get(hashes)
        else:
            data = storage.content_get_data(hashes)
        if data is None:
            content = SKIPPED_MESSAGE
        else:
            content = data
    elif status == "absent":
        content = SKIPPED_MESSAGE
    elif status == "hidden":
        content = HIDDEN_MESSAGE
    elif status is None:
        content = MISSING_MESSAGE
    else:
        assert False, (
            f"unexpected status {status!r} "
            f"for content {hashutil.hash_to_hex(file_data['target'])}"
        )

    return {"content": content, **file_data}


class DirectoryBuilder:
    """Reconstructs the on-disk representation of a directory in the storage."""

    def __init__(
        self,
        storage: StorageInterface,
        root: bytes,
        dir_id: bytes,
        thread_pool_size: int = 10,
        objstorage: Optional[ObjStorageInterface] = None,
    ):
        """Initialize the directory builder.

        Args:
            storage: the storage object
            root: the path where the directory should be reconstructed
            dir_id: the identifier of the directory in the storage
        """
        self.storage = storage
        self.root = root
        self.dir_id = dir_id
        self.thread_pool_size = thread_pool_size
        self.objstorage = objstorage

    def build(self) -> None:
        """Perform the reconstruction of the directory in the given root."""

        def file_fetcher(file_data: Dict[str, Any]) -> None:
            file_data = get_filtered_file_content(
                self.storage, file_data, self.objstorage
            )
            path = os.path.join(self.root, file_data["path"])
            self._create_file(path, file_data["content"], file_data["perms"])

        executor = concurrent.futures.ThreadPoolExecutor(self.thread_pool_size)
        futures = []

        os.makedirs(self.root, exist_ok=True)
        queue = collections.deque([(b"", self.dir_id)])
        while queue:
            path, dir_id = queue.popleft()
            dir_entries = self.storage.directory_ls(dir_id)
            for dir_entry in dir_entries:
                dir_entry["path"] = os.path.join(path, dir_entry["name"])
                match dir_entry["type"]:
                    case "dir":
                        self._create_tree(dir_entry)
                        queue.append((dir_entry["path"], dir_entry["target"]))
                    case "rev":
                        self._create_revision(dir_entry)
                    case "file":
                        futures.append(executor.submit(file_fetcher, dir_entry))
                    case _:
                        raise ValueError(
                            f"Unsupported directory entry type {dir_entry['type']} for "
                            f"{dir_entry['name']:r} in directory swh:1:dir:{dir_id.hex()}"
                        )

        concurrent.futures.wait(futures)

    def _create_tree(self, directory: Dict[str, Any]) -> None:
        """Create a directory tree from root for the given path."""
        os.makedirs(os.path.join(self.root, directory["path"]), exist_ok=True)

    def _create_revision(self, rev_data: Dict[str, Any]) -> None:
        """Create the revision in the tree as a broken symlink to the target
        identifier."""
        os.makedirs(os.path.join(self.root, rev_data["path"]), exist_ok=True)

    def _create_file(
        self, path: bytes, content: bytes, mode: int = DentryPerms.content
    ) -> None:
        """Create the given file and fill it with content."""
        perms = mode_to_perms(mode)
        if perms == DentryPerms.symlink:
            os.symlink(content, path)
        else:
            with open(path, "wb") as f:
                f.write(content)
            os.chmod(path, perms.value)



# Copyright (C) 2018-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import textwrap
from unittest.mock import MagicMock

from swh.model.swhids import CoreSWHID
from swh.vault.cookers.base import BaseVaultCooker

TEST_BUNDLE_CHUNKS = [b"test content 1\n", b"test content 2\n", b"test content 3\n"]
TEST_BUNDLE_CONTENT = b"".join(TEST_BUNDLE_CHUNKS)
TEST_BUNDLE_TYPE = "test_type"
TEST_SWHID = CoreSWHID.from_string("swh:1:cnt:17a3e48bce37be5226490e750202ad3a9a1a3fe9")


class BaseVaultCookerMock(BaseVaultCooker):
    BUNDLE_TYPE = TEST_BUNDLE_TYPE

    def __init__(self):
        # we do not call super() here to bypass the building of db objects from
        # config since we do mock these db objects
        self.config = {}
        self.storage = MagicMock()
        self.backend = MagicMock()
        self.swhid = TEST_SWHID
        self.obj_id = TEST_SWHID.object_id
        self.max_bundle_size = 1024

    def check_exists(self):
        return True

    def prepare_bundle(self):
        for chunk in TEST_BUNDLE_CHUNKS:
            self.write(chunk)


def test_simple_cook():
    cooker = BaseVaultCookerMock()
    cooker.cook()
    cooker.backend.put_bundle.assert_called_once_with(
        TEST_BUNDLE_TYPE, TEST_SWHID, TEST_BUNDLE_CONTENT
    )
    cooker.backend.set_status.assert_called_with(TEST_BUNDLE_TYPE, TEST_SWHID, "done")
    cooker.backend.set_progress.assert_called_with(TEST_BUNDLE_TYPE, TEST_SWHID, None)
    cooker.backend.send_notif.assert_called_with(TEST_BUNDLE_TYPE, TEST_SWHID)


def test_code_exception_cook():
    cooker = BaseVaultCookerMock()
    cooker.prepare_bundle = MagicMock()
    cooker.prepare_bundle.side_effect = RuntimeError("Nope")
    cooker.cook()

    # Potentially remove this when we have objstorage streaming
    cooker.backend.put_bundle.assert_not_called()

    cooker.backend.set_status.assert_called_with(TEST_BUNDLE_TYPE, TEST_SWHID, "failed")
    assert cooker.backend.set_progress.call_args[0][2].startswith(
        textwrap.dedent(
            """\
            Internal Server Error. This incident will be reported.
            The full error was:

            Traceback (most recent call last):
            """
        )
    )
    cooker.backend.send_notif.assert_called_with(TEST_BUNDLE_TYPE, TEST_SWHID)


def test_policy_exception_cook():
    cooker = BaseVaultCookerMock()
    cooker.max_bundle_size = 8
    cooker.cook()

    # Potentially remove this when we have objstorage streaming
    cooker.backend.put_bundle.assert_not_called()

    cooker.backend.set_status.assert_called_with(TEST_BUNDLE_TYPE, TEST_SWHID, "failed")
    assert "exceeds" in cooker.backend.set_progress.call_args[0][2]
    cooker.backend.send_notif.assert_called_with(TEST_BUNDLE_TYPE, TEST_SWHID)



# Copyright (C) 2017-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
import os
import time
import zlib

from fastimport.commands import (
    BlobCommand,
    CommitCommand,
    FileDeleteCommand,
    FileModifyCommand,
    ResetCommand,
)

from swh.model import hashutil
from swh.model.from_disk import DentryPerms, mode_to_perms
from swh.model.swhids import ObjectType
from swh.model.toposort import toposort
from swh.vault.cookers.base import BaseVaultCooker
from swh.vault.cookers.utils import revision_log
from swh.vault.to_disk import get_filtered_file_content


class RevisionGitfastCooker(BaseVaultCooker):
    """Cooker to create a git fast-import bundle"""

    BUNDLE_TYPE = "gitfast"
    SUPPORTED_OBJECT_TYPES = {ObjectType.REVISION}

    def check_exists(self):
        return not list(self.storage.revision_missing([self.obj_id]))

    def prepare_bundle(self):
        self.log = list(toposort(revision_log(self.storage, self.obj_id)))
        self.gzobj = zlib.compressobj(9, zlib.DEFLATED, zlib.MAX_WBITS | 16)
        self.fastexport()
        self.write(self.gzobj.flush())

    def write_cmd(self, cmd):
        chunk = bytes(cmd) + b"\n"
        super().write(self.gzobj.compress(chunk))

    def fastexport(self):
        """Generate all the git fast-import commands from a given log."""
        self.rev_by_id = {r["id"]: r for r in self.log}
        self.obj_done = set()
        self.obj_to_mark = {}
        self.next_available_mark = 1

        last_progress_report = None

        for i, rev in enumerate(self.log, 1):
            # Update progress if needed
            ct = time.time()
            if last_progress_report is None or last_progress_report + 2 <= ct:
                last_progress_report = ct
                pg = "Computing revision {}/{}".format(i, len(self.log))
                self.backend.set_progress(self.BUNDLE_TYPE, self.swhid, pg)

            # Compute the current commit
            self._compute_commit_command(rev)

    def mark(self, obj_id):
        """Get the mark ID as bytes of a git object.

        If the object has not yet been marked, assign a new ID and add it to
        the mark dictionary.
        """
        if obj_id not in self.obj_to_mark:
            self.obj_to_mark[obj_id] = self.next_available_mark
            self.next_available_mark += 1
        return str(self.obj_to_mark[obj_id]).encode()

    def _compute_blob_command_content(self, file_data):
        """Compute the blob command of a file entry if it has not been
        computed yet.
        """
        obj_id = file_data["sha1"]
        if obj_id in self.obj_done:
            return
        content = get_filtered_file_content(self.storage, file_data)
        self.write_cmd(BlobCommand(mark=self.mark(obj_id), data=content["content"]))
        self.obj_done.add(obj_id)

    def _author_tuple_format(self, author, date):
        # We never want to have None values here so we replace null entries
        # by ''.
        if author is not None:
            author_tuple = (author.get("name") or b"", author.get("email") or b"")
        else:
            author_tuple = (b"", b"")
        if date is not None:
            date_tuple = (
                date.get("timestamp", {}).get("seconds") or 0,
                (date.get("offset") or 0) * 60,
            )
        else:
            date_tuple = (0, 0)
        return author_tuple + date_tuple

    def _compute_commit_command(self, rev):
        """Compute a commit command from a specific revision."""
        if "parents" in rev and rev["parents"]:
            from_ = b":" + self.mark(rev["parents"][0])
            merges = [b":" + self.mark(r) for r in rev["parents"][1:]]
            parent = self.rev_by_id[rev["parents"][0]]
        else:
            # We issue a reset command before all the new roots so that they
            # are not automatically added as children of the current branch.
            self.write_cmd(ResetCommand(b"refs/heads/master", None))
            from_ = None
            merges = None
            parent = None

        # Retrieve the file commands while yielding new blob commands if
        # needed.
        files = list(self._compute_file_commands(rev, parent))

        # Construct and write the commit command
        author = self._author_tuple_format(rev["author"], rev["date"])
        committer = self._author_tuple_format(rev["committer"], rev["committer_date"])
        self.write_cmd(
            CommitCommand(
                ref=b"refs/heads/master",
                mark=self.mark(rev["id"]),
                author=author,
                committer=committer,
                message=rev["message"] or b"",
                from_=from_,
                merges=merges,
                file_iter=files,
            )
        )

    @functools.lru_cache(maxsize=4096)
    def _get_dir_ents(self, dir_id=None):
        """Get the entities of a directory as a dictionary (name -> entity).

        This function has a cache to avoid doing multiple requests to retrieve
        the same entities, as doing a directory_ls() is expensive.
        """
        data = self.storage.directory_ls(dir_id) if dir_id is not None else []
        return {f["name"]: f for f in data}

    def _compute_file_commands(self, rev, parent=None):
        """Compute all the file commands of a revision.

        Generate a diff of the files between the revision and its main parent
        to find the necessary file commands to apply.
        """
        # Initialize the stack with the root of the tree.
        cur_dir = rev["directory"]
        parent_dir = parent["directory"] if parent else None
        stack = [(b"", cur_dir, parent_dir)]

        while stack:
            # Retrieve the current directory and the directory of the parent
            # commit in order to compute the diff of the trees.
            root, cur_dir_id, prev_dir_id = stack.pop()
            cur_dir = self._get_dir_ents(cur_dir_id)
            prev_dir = self._get_dir_ents(prev_dir_id)

            # Find subtrees to delete:
            #  - Subtrees that are not in the new tree (file or directory
            #    deleted).
            #  - Subtrees that do not have the same type in the new tree
            #    (file -> directory or directory -> file)
            # After this step, every node remaining in the previous directory
            # has the same type than the one in the current directory.
            for fname, f in prev_dir.items():
                if fname not in cur_dir or f["type"] != cur_dir[fname]["type"]:
                    yield FileDeleteCommand(path=os.path.join(root, fname))

            # Find subtrees to modify:
            #  - Leaves (files) will be added or modified using `filemodify`
            #  - Other subtrees (directories) will be added to the stack and
            #    processed in the next iteration.
            for fname, f in cur_dir.items():
                # A file is added or modified if it was not in the tree, if its
                # permissions changed or if its content changed.
                if f["type"] == "file" and (
                    fname not in prev_dir
                    or f["sha1"] != prev_dir[fname]["sha1"]
                    or f["perms"] != prev_dir[fname]["perms"]
                ):
                    # Issue a blob command for the new blobs if needed.
                    self._compute_blob_command_content(f)
                    yield FileModifyCommand(
                        path=os.path.join(root, fname),
                        mode=mode_to_perms(f["perms"]).value,
                        dataref=(b":" + self.mark(f["sha1"])),
                        data=None,
                    )
                # A revision is added or modified if it was not in the tree or
                # if its target changed
                elif f["type"] == "rev" and (
                    fname not in prev_dir or f["target"] != prev_dir[fname]["target"]
                ):
                    yield FileModifyCommand(
                        path=os.path.join(root, fname),
                        mode=DentryPerms.revision,
                        dataref=hashutil.hash_to_hex(f["target"]).encode(),
                        data=None,
                    )
                # A directory is added or modified if it was not in the tree or
                # if its target changed.
                elif f["type"] == "dir":
                    f_prev_target = None
                    if fname in prev_dir and prev_dir[fname]["type"] == "dir":
                        f_prev_target = prev_dir[fname]["target"]
                    if f_prev_target is None or f["target"] != f_prev_target:
                        stack.append(
                            (os.path.join(root, fname), f["target"], f_prev_target)
                        )



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Callable, Dict, List, Tuple

from swh.model.swhids import CoreSWHID

ENCODERS: List[Tuple[type, str, Callable]] = [
    (CoreSWHID, "core_swhid", str),
]


DECODERS: Dict[str, Callable] = {
    "core_swhid": CoreSWHID.from_string,
}
